#!/usr/bin/env python
from django.conf import settings
from django.core.management import call_command


if not settings.configured:
    settings.configure(
        DATABASES = {
            'default': {
                'ENGINE': 'django.db.backends.sqlite3',
                'NAME': ':memory:',
            },
            'remote_db': {
                'ENGINE': 'django.db.backends.sqlite3',
                'NAME': ':memory:',
            }
        },
        INSTALLED_APPS = (
            'django.contrib.admin',
            'django.contrib.auth',
            'django.contrib.contenttypes',
            'django.contrib.sites',
            'django.contrib.sessions',
            'dbsettings',
            'synchro',
        ),
        SITE_ID = 1,
        SYNCHRO_REMOTE = 'remote_db',
        ROOT_URLCONF = 'can be anything - tests override this',
    )

call_command('test', 'synchro')



from datetime import datetime

from django.core.exceptions import MultipleObjectsReturned, ValidationError
from django.db.models import Manager, Model
from django.db.models.base import ModelBase


class NaturalManager(Manager):
    """
    Manager must be able to instantiate without arguments in order to work with M2M.
    Hence this machinery to store arguments in class.
    Somehow related to Django bug #13313.
    """
    allow_many = False

    def get_by_natural_key(self, *args):
        lookups = dict(zip(self.fields, args))
        try:
            return self.get(**lookups)
        except MultipleObjectsReturned:
            if self.allow_many:
                return self.filter(**lookups)[0]
            raise

    def __new__(cls, *fields, **options):
        """
        Creates actual manager, which can be further subclassed and instantiated without arguments.
        """
        if not fields and hasattr(cls, 'fields') and hasattr(cls, 'allow_many'):
            # Class was already prepared.
            return super(NaturalManager, cls).__new__(cls)

        assert fields, 'No fields specified in %s constructor' % cls
        _fields = fields
        _allow_many = options.get('allow_many', False)
        manager = options.get('manager', Manager)
        if not issubclass(manager, Manager):
            raise ValidationError(
                '%s manager class must be a subclass of django.db.models.Manager.'
                % manager.__name__)

        class NewNaturalManager(cls, manager):
            fields = _fields
            allow_many = _allow_many

            def __init__(self, *args, **kwargs):
                # Intentionally ignore arguments
                super(NewNaturalManager, self).__init__()
        return super(NaturalManager, cls).__new__(NewNaturalManager)


class _NaturalKeyModelBase(ModelBase):
    def __new__(cls, name, bases, attrs):
        parents = [b for b in bases if isinstance(b, _NaturalKeyModelBase)]
        if not parents:
            return super(_NaturalKeyModelBase, cls).__new__(cls, name, bases, attrs)
        kwargs = {}
        if 'objects' in attrs:
            kwargs['manager'] = attrs['objects'].__class__
        kwargs.update(attrs.pop('_natural_manager_kwargs', {}))
        attrs['objects'] = NaturalManager(*attrs['_natural_key'], **kwargs)
        return super(_NaturalKeyModelBase, cls).__new__(cls, name, bases, attrs)


class NaturalKeyModel(Model):
    __metaclass__ = _NaturalKeyModelBase
    _natural_key = ()

    def natural_key(self):
        return tuple(getattr(self, field) for field in self._natural_key)

    class Meta:
        abstract = True


def reset_synchro():
    from models import ChangeLog, Reference, options
    options.last_check = datetime.now()
    ChangeLog.objects.all().delete()
    Reference.objects.all().delete()



from django.contrib import admin
from django.conf.urls import patterns, url, include


urlpatterns = patterns('',
    url(r'^admin/', include(admin.site.urls)),
    url(r'^synchro/', include('synchro.urls')),
)



import datetime

from django import get_version
from django.conf import settings
from django.core.exceptions import ValidationError, ImproperlyConfigured
from django.core.management import call_command, CommandError
from django.core.urlresolvers import reverse
from django.db import models
from django.db.models import F
from django.db.models.signals import pre_save, post_save, post_delete
from django.dispatch import receiver
from django.test import TestCase
from django.test.utils import override_settings
from django.utils.unittest.case import skipUnless

from models import ChangeLog
import settings as synchro_settings
from signals import DisableSynchroLog, disable_synchro_log
from utility import NaturalManager, reset_synchro, NaturalKeyModel

if get_version() >= '1.5':
    from django.contrib.auth import get_user_model
    User = get_user_model()

    def user_model_quite_standard():
        "Check if installed User object is not too custom for the tests to instantiate it."
        from django.contrib.auth.models import User as StandardUser
        if (User.USERNAME_FIELD == StandardUser.USERNAME_FIELD and
            User.REQUIRED_FIELDS == StandardUser.REQUIRED_FIELDS):
            return True
        return False
else:
    from django.contrib.auth.models import User
    user_model_quite_standard = lambda: True

LOCAL = 'default'
REMOTE = settings.SYNCHRO_REMOTE
# List of test models
SETTINGS = {
    'SYNCHRO_MODELS': (
        ('synchro', 'testmodel', 'PkModelWithSkip', 'ModelWithKey', 'ModelWithFK', 'A', 'X',
         'M2mModelWithKey', 'M2mAnother', 'M2mModelWithInter', 'M2mSelf'),
    )
}


def contrib_apps(*apps):
    """Check if all listed apps are installed."""
    for app in apps:
        if 'django.contrib.%s' % app not in settings.INSTALLED_APPS:
            return False
    return True


##### Test models ################################


class TestModel(models.Model):
    name = models.CharField(max_length=10)
    cash = models.IntegerField(default=0)


class PkModelWithSkip(models.Model):
    name = models.CharField(max_length=10, primary_key=True)
    cash = models.IntegerField(default=0)
    visits = models.PositiveIntegerField(default=0)
    SYNCHRO_SKIP = ('visits',)


class ModelWithFK(models.Model):
    name = models.CharField(max_length=10)
    visits = models.PositiveIntegerField(default=0)
    link = models.ForeignKey(PkModelWithSkip, related_name='links')


@receiver(pre_save, sender=ModelWithFK)
def save_prev(sender, instance, **kwargs):
    """Save object's previous state (before save)."""
    try:
        instance._prev = sender.objects.db_manager(instance._state.db).get(pk=instance.pk)
    except sender.DoesNotExist:
        instance._prev = None


@receiver(post_save, sender=ModelWithFK)
@disable_synchro_log
def update_visits(sender, instance, created, **kwargs):
    """Update parent visits."""
    if not created:
        # Side note: in the statement below it should be instance._prev.link in case of link change,
        # but it requires some refreshing from database (since instance._prev.link and instance.link
        # are two different instances of the same object). For this test
        instance.link.visits -= instance._prev.visits
        instance.link.save()
    instance.link.visits += instance.visits
    instance.link.save()


class CustomManager(models.Manager):
    def foo(self):
        return 'bar'

    def none(self):  # Overrides Manager method
        return 'Not a single object!'


class MyNaturalManager(NaturalManager, CustomManager):
    fields = ('name',)


class ModelWithKey(NaturalKeyModel):
    name = models.CharField(max_length=10)
    cash = models.IntegerField(default=0)
    visits = models.PositiveIntegerField(default=0)
    SYNCHRO_SKIP = ('visits',)
    _natural_key = ('name',)

    objects = CustomManager()
    another_objects = MyNaturalManager()


class M2mModelWithKey(models.Model):
    foo = models.IntegerField(default=1)
    objects = NaturalManager('foo')

    def natural_key(self):
        return self.foo,


class M2mAnother(models.Model):
    bar = models.IntegerField(default=1)
    m2m = models.ManyToManyField('M2mModelWithKey', related_name='r_m2m')


class M2mModelWithInter(models.Model):
    bar = models.IntegerField(default=1)
    m2m = models.ManyToManyField('M2mModelWithKey', related_name='r_m2m_i',
                                 through='M2mIntermediate')


class M2mNotExplicitlySynced(models.Model):
    # This model is not listed in SYNCHRO_MODELS
    foo = models.IntegerField(default=1)


class M2mIntermediate(models.Model):
    with_key = models.ForeignKey(M2mModelWithKey)
    with_inter = models.ForeignKey(M2mModelWithInter)
    # To get everything worse, use another FK here, in order to test intermediate sync.
    extra = models.ForeignKey(M2mNotExplicitlySynced)
    cash = models.IntegerField()


class M2mSelf(models.Model):
    foo = models.IntegerField(default=1)
    m2m = models.ManyToManyField('self')


class A(models.Model):
    foo = models.IntegerField(default=1)
    bar = models.IntegerField(default=1)


class X(models.Model):
    name = models.CharField(max_length=10)


def update_bar_bad(sender, using, **kwargs):
    a = A.objects.db_manager(using).all()[0]
    a.bar += 1
    a.save()


@disable_synchro_log
def update_bar_good_dis(sender, using, **kwargs):
    a = A.objects.db_manager(using).all()[0]
    a.bar += 1
    a.save()


def update_bar_good_upd(sender, using, **kwargs):
    A.objects.db_manager(using).update(bar=F('bar') + 1)  # update don't emmit signals


##### Tests themselves ###########################


@override_settings(**SETTINGS)
class SynchroTests(TestCase):
    multi_db = True
    urls = 'synchro.test_urls'

    @classmethod
    def setUpClass(cls):
        """Update SYNCHRO_MODELS and reload them"""
        with override_settings(**SETTINGS):
            reload(synchro_settings)

    @classmethod
    def tearDownClass(cls):
        """Clean up after yourself: restore the previous SYNCHRO_MODELS"""
        reload(synchro_settings)

    def _assertDbCount(self, db, num, cls):
        self.assertEqual(num, cls.objects.db_manager(db).count())

    def assertLocalCount(self, num, cls):
        self._assertDbCount(LOCAL, num, cls)

    def assertRemoteCount(self, num, cls):
        self._assertDbCount(REMOTE, num, cls)

    def synchronize(self, **kwargs):
        call_command('synchronize', verbosity=0, **kwargs)

    def wait(self):
        """
        Since tests are run too fast, we need to wait for a moment, so that some ChangeLog objects
        could be considered "old" and not synchronized again.
        Waiting one second every time this method is called would lengthen tests - so instead we
        simulate time shift.
        """
        ChangeLog.objects.update(date=F('date') - datetime.timedelta(seconds=1))
        ChangeLog.objects.db_manager(REMOTE).update(date=F('date') - datetime.timedelta(seconds=1))

    def reset(self):
        reset_synchro()

    def assertNoActionOnSynchronize(self, sender, save=True, delete=True):
        def fail(**kwargs):
            self.fail('Signal caught - action performed.')
        if save:
            post_save.connect(fail, sender=sender)
        if delete:
            post_delete.connect(fail, sender=sender)
        self.synchronize()
        post_save.disconnect(fail, sender=sender)
        post_delete.disconnect(fail, sender=sender)


class SimpleSynchroTests(SynchroTests):
    """Cover basic functionality."""

    def test_settings(self):
        """Check if test SYNCHRO_MODELS is loaded."""
        self.assertIn(TestModel, synchro_settings.MODELS)
        self.assertIn(PkModelWithSkip, synchro_settings.MODELS)
        self.assertNotIn(ChangeLog, synchro_settings.MODELS)

    def test_app_paths(self):
        """Check if app in SYNCHRO_MODELS can be stated in any way."""
        from django.contrib.auth.models import Group
        self.assertNotIn(Group, synchro_settings.MODELS)

        INSTALLED_APPS = settings.INSTALLED_APPS + ('django.contrib.auth',)
        with override_settings(INSTALLED_APPS=INSTALLED_APPS):
            # fully qualified path
            with override_settings(SYNCHRO_MODELS=('django.contrib.auth',)):
                reload(synchro_settings)
                self.assertIn(Group, synchro_settings.MODELS)
            # app label
            with override_settings(SYNCHRO_MODELS=('auth',)):
                reload(synchro_settings)
                self.assertIn(Group, synchro_settings.MODELS)

        # Restore previous state
        reload(synchro_settings)
        self.assertNotIn(Group, synchro_settings.MODELS)

    def test_settings_with_invalid_remote(self):
        """Check if specifying invalid remote results in exception."""
        with override_settings(SYNCHRO_REMOTE='invalid'):
            with self.assertRaises(ImproperlyConfigured):
                reload(synchro_settings)
        # Restore previous state
        reload(synchro_settings)
        self.assertEqual(REMOTE, synchro_settings.REMOTE)

    def test_settings_without_remote(self):
        """Check if lack of REMOTE in settings cause synchronization disablement."""
        import synchro.management.commands.synchronize
        with override_settings(SYNCHRO_REMOTE=None):
            reload(synchro_settings)
            reload(synchro.management.commands.synchronize)
            self.assertIsNone(synchro_settings.REMOTE)
            self.assertLocalCount(0, ChangeLog)
            TestModel.objects.create(name='James', cash=7)
            self.assertLocalCount(1, TestModel)
            # ChangeLog created successfully despite lack of REMOTE
            self.assertLocalCount(1, ChangeLog)

            from django import get_version
            if get_version() >= '1.5':
                self.assertRaises(CommandError, self.synchronize)
            else:
                # Because of BaseCommand bug (#18387, fixed in Django 1.5), we cannot use
                # CommandError in tests: it is being caught and results in sys.exit(1).
                class ExceptionStub(Exception):
                    pass
                with self.assertRaises(ExceptionStub):
                    self.synchronize(exception_class=ExceptionStub)

        # Restore previous state
        reload(synchro_settings)
        reload(synchro.management.commands.synchronize)
        self.assertEqual(REMOTE, synchro_settings.REMOTE)

    def test_simple_synchro(self):
        """Check object creation and checkpoint storage."""
        prev = datetime.datetime.now()
        a = TestModel.objects.create(name='James', cash=7)
        self.assertLocalCount(1, TestModel)
        self.assertRemoteCount(0, TestModel)
        self.synchronize()
        self.assertLocalCount(1, TestModel)
        self.assertRemoteCount(1, TestModel)
        b = TestModel.objects.db_manager(REMOTE).all()[0]
        self.assertFalse(a is b)
        self.assertEqual(a.name, b.name)
        self.assertEqual(a.cash, b.cash)
        from synchro.models import options
        self.assertTrue(options.last_check >= prev.replace(microsecond=0))

    def test_auto_pk(self):
        """
        Test if auto pk is *not* overwritten.
        Although local object has the same pk as remote one, new object will be created,
        because pk is automatic.
        """
        some = TestModel.objects.db_manager(REMOTE).create(name='Remote James', cash=77)
        a = TestModel.objects.create(name='James', cash=7)
        self.assertEquals(a.pk, some.pk)
        self.synchronize()
        self.assertLocalCount(1, TestModel)
        self.assertRemoteCount(2, TestModel)
        self.assertTrue(TestModel.objects.db_manager(REMOTE).get(name='James'))
        self.assertTrue(TestModel.objects.db_manager(REMOTE).get(name='Remote James'))

    def test_not_auto_pk(self):
        """
        Test if explicit pk *is overwritten*.
        If local object has the same pk as remote one, remote object will be completely overwritten.
        """
        some = PkModelWithSkip.objects.db_manager(REMOTE).create(name='James', cash=77, visits=5)
        a = PkModelWithSkip.objects.create(name='James', cash=7, visits=42)
        self.assertEquals(a.pk, some.pk)
        self.synchronize()
        self.assertLocalCount(1, PkModelWithSkip)
        self.assertRemoteCount(1, PkModelWithSkip)
        b = PkModelWithSkip.objects.db_manager(REMOTE).get(name='James')
        self.assertEqual(7, b.cash)
        # Because whole object is copied, skipping use default value.
        self.assertEqual(0, b.visits)

    def test_change(self):
        """Test simple change"""
        a = TestModel.objects.create(name='James', cash=7)
        self.synchronize()
        self.wait()
        a.name = 'Bond'
        a.save()
        self.synchronize()
        b = TestModel.objects.db_manager(REMOTE).get(cash=7)
        self.assertEqual(a.name, b.name)

    def test_skipping_add(self):
        """Test if field is skipped during creation - that is, cleared."""
        PkModelWithSkip.objects.create(name='James', cash=7, visits=42)
        self.synchronize()
        b = PkModelWithSkip.objects.db_manager(REMOTE).get(name='James')
        self.assertEqual(7, b.cash)
        self.assertEqual(0, b.visits)  # Skipping use default value when creating

    def test_skipping_change(self):
        """Test if field is skipped."""
        a = PkModelWithSkip.objects.create(name='James', cash=7)
        self.synchronize()
        b = PkModelWithSkip.objects.db_manager(REMOTE).get(name='James')
        b.visits = 42
        b.save()
        self.wait()
        a.cash = 77
        a.save()
        self.synchronize()
        b = PkModelWithSkip.objects.db_manager(REMOTE).get(name='James')
        self.assertEqual(a.cash, b.cash)
        self.assertEqual(42, b.visits)

    def test_deletion(self):
        """Test deletion."""
        a = TestModel.objects.create(name='James', cash=7)
        self.synchronize()
        self.assertRemoteCount(1, TestModel)
        self.wait()
        a.delete()
        self.synchronize()
        self.assertRemoteCount(0, TestModel)

    def test_untracked_deletion(self):
        """Test if deletion is not performed on lack of Reference and key."""
        TestModel.objects.db_manager(REMOTE).create(name='James', cash=7)
        a = TestModel.objects.create(name='James', cash=7)
        self.reset()
        a.delete()
        self.synchronize()
        self.assertLocalCount(0, TestModel)
        self.assertRemoteCount(1, TestModel)

    def test_add_del(self):
        """Test if no unnecessary action is performed if added and deleted."""
        a = TestModel.objects.create(name='James')
        a.delete()
        self.assertNoActionOnSynchronize(TestModel)
        self.assertRemoteCount(0, TestModel)

    def test_chg_del(self):
        """Test if no unnecessary action is performed if changed and deleted."""
        a = TestModel.objects.create(name='James', cash=7)
        self.synchronize()
        self.wait()
        a.name = 'Bond'
        a.save()
        a.delete()
        self.assertNoActionOnSynchronize(TestModel, delete=False)
        self.assertRemoteCount(0, TestModel)

    def test_add_chg_del_add_chg(self):
        """Combo."""
        a = TestModel.objects.create(name='James', cash=7)
        a.name = 'Bond'
        a.save()
        a.delete()
        a = TestModel.objects.create(name='Vimes', cash=7)
        a.cash = 77
        a.save()
        self.synchronize()
        self.assertRemoteCount(1, TestModel)
        b = TestModel.objects.db_manager(REMOTE).get(name='Vimes')
        self.assertEqual(a.cash, b.cash)

    def test_reference(self):
        """Test if object once synchronized is linked with remote instance."""
        some = TestModel.objects.db_manager(REMOTE).create(name='Remote James', cash=77)
        a = TestModel.objects.create(name='James', cash=7)
        self.assertEquals(a.pk, some.pk)
        self.synchronize()
        self.assertRemoteCount(2, TestModel)
        b = TestModel.objects.db_manager(REMOTE).get(name='James')
        self.assertNotEquals(a.pk, b.pk)
        b.name = 'Bond'
        b.save()  # This change will be discarded
        self.wait()
        a.cash = 42
        a.save()
        self.synchronize()
        b = TestModel.objects.db_manager(REMOTE).get(pk=b.pk)
        self.assertEqual(a.name, b.name)
        self.assertEqual(a.cash, b.cash)

    def test_time_comparing(self):
        """Test if synchronization is not performed if REMOTE object is newer."""
        a = TestModel.objects.create(name="James", cash=7)
        self.synchronize()
        self.assertRemoteCount(1, TestModel)
        self.wait()
        a.cash = 42  # local change
        a.save()
        self.wait()
        b = TestModel.objects.db_manager(REMOTE).get(name="James")
        b.cash = 77  # remote change, done later
        b.save()
        self.assertNoActionOnSynchronize(TestModel)
        self.assertRemoteCount(1, TestModel)
        b = TestModel.objects.db_manager(REMOTE).get(name="James")
        self.assertEqual(77, b.cash)  # remote object hasn't changed

    @skipUnless(contrib_apps('admin', 'auth', 'sessions'),
                'admin, auth or sessions not in INSTALLED_APPS')
    @skipUnless(user_model_quite_standard(), 'Too custom User model')
    def test_admin(self):
        """Test if synchronization can be performed via admin interface."""
        path = reverse('synchro')
        user = User._default_manager.create_user('admin', 'mail', 'admin')
        self.client.login(username='admin', password='admin')
        # test if staff status is required
        self.assertTemplateUsed(self.client.get(path), 'admin/login.html')
        user.is_staff = True
        user.save()
        # superuser
        self.assertTemplateUsed(self.client.get(path), 'synchro.html')
        # actual synchronization
        self.reset()
        TestModel.objects.create(name='James', cash=7)
        self.assertRemoteCount(0, TestModel)
        self.client.post(path, {'synchro': True})  # button clicked
        self.assertRemoteCount(1, TestModel)


class AdvancedSynchroTests(SynchroTests):
    """Cover additional features."""

    def test_manager_class(self):
        """Test if NaturalManager works."""
        self.assertIsInstance(ModelWithKey.objects, NaturalManager)
        self.assertIsInstance(ModelWithKey.another_objects, NaturalManager)
        # Test if it subclasses user manager as well
        self.assertIsInstance(ModelWithKey.objects, CustomManager)
        self.assertIsInstance(ModelWithKey.another_objects, CustomManager)
        self.assertEqual('bar', ModelWithKey.objects.foo())
        self.assertEqual('bar', ModelWithKey.another_objects.foo())
        # Check proper MRO: NaturalManager, user manager, Manager
        self.assertTrue(hasattr(ModelWithKey.objects, 'get_by_natural_key'))
        self.assertTrue(hasattr(ModelWithKey.another_objects, 'get_by_natural_key'))
        self.assertEqual('Not a single object!', ModelWithKey.objects.none())
        self.assertEqual('Not a single object!', ModelWithKey.another_objects.none())
        self.assertSequenceEqual([], ModelWithKey.objects.all())
        self.assertSequenceEqual([], ModelWithKey.another_objects.all())

        # Test get_by_natural_key
        obj = ModelWithKey.objects.create(name='James')
        self.assertEqual(obj.pk, ModelWithKey.objects.get_by_natural_key('James').pk)
        self.assertEqual(obj.pk, ModelWithKey.another_objects.get_by_natural_key('James').pk)

        # Test instantiating (DJango #13313: manager must be instantiable without arguments)
        try:
            ModelWithKey.objects.__class__()
            ModelWithKey.another_objects.__class__()
        except TypeError:
            self.fail('Cannot instantiate.')

        # Test if class checking occurs
        def wrong():
            class BadManager:
                pass

            class X(models.Model):
                x = models.IntegerField()
                objects = NaturalManager('x', manager=BadManager)
        self.assertRaises(ValidationError, wrong)  # User manager must subclass Manager

        # Test if manager without fields raises exception
        def wrong2():
            class X(models.Model):
                x = models.IntegerField()
                objects = NaturalManager()
        self.assertRaises(AssertionError, wrong2)

    def test_natural_key(self):
        """
        Test if natural key works.
        If local object has the same key as remote one, remote object will be updated.
        """
        b = ModelWithKey.objects.db_manager(REMOTE).create(name='James', cash=77, visits=5)
        a = ModelWithKey.objects.create(name='James', cash=7, visits=42, pk=2)
        self.assertNotEquals(a.pk, b.pk)
        self.synchronize()
        self.assertLocalCount(1, ModelWithKey)
        self.assertRemoteCount(1, ModelWithKey)
        remote = ModelWithKey.objects.db_manager(REMOTE).get(name='James')
        self.assertEqual(7, remote.cash)
        # Because remote object is found, skipping use remote value (not default).
        self.assertEqual(5, remote.visits)

    def test_natural_key_deletion(self):
        """
        Test if natural key works on deletion.
        When no Reference exist, delete object matching natural key.
        """
        ModelWithKey.objects.db_manager(REMOTE).create(name='James', cash=77, visits=5)
        a = ModelWithKey.objects.create(name='James', cash=7, visits=42, pk=2)
        self.reset()
        a.delete()
        self.synchronize()
        self.assertLocalCount(0, ModelWithKey)
        self.assertRemoteCount(0, ModelWithKey)

    def test_foreign_keys(self):
        """Test if foreign keys are synchronized."""
        a = PkModelWithSkip.objects.create(name='James')
        self.reset()  # Even if parent model is not recorded!
        ModelWithFK.objects.create(name='1', link=a)
        ModelWithFK.objects.create(name='2', link=a)
        self.synchronize()
        self.assertRemoteCount(1, PkModelWithSkip)
        self.assertRemoteCount(2, ModelWithFK)
        b = PkModelWithSkip.objects.db_manager(REMOTE).get(name='James')
        self.assertEqual(2, b.links.count())
        # Check if all submodels belong to remote db
        self.assertTrue(all(map(lambda x: x._state.db == REMOTE, b.links.all())))

    def test_disabling(self):
        """Test if logging can be disabled."""
        # with context
        with DisableSynchroLog():
            TestModel.objects.create(name='James')
        self.synchronize()
        self.assertLocalCount(1, TestModel)
        self.assertRemoteCount(0, TestModel)

        # with decorator
        @disable_synchro_log
        def create():
            PkModelWithSkip.objects.create(name='James')
        create()
        self.synchronize()
        self.assertLocalCount(1, PkModelWithSkip)
        self.assertRemoteCount(0, PkModelWithSkip)


class SignalSynchroTests(SynchroTests):
    """Cover signals tests."""

    def test_signals_and_skip(self):
        """Some signal case from real life."""
        a = PkModelWithSkip.objects.create(name='James', cash=7)
        self.synchronize()
        self.wait()
        b = PkModelWithSkip.objects.db_manager(REMOTE).get(name='James')
        b.cash = 77  # some remote changes
        b.visits = 10
        b.save()
        self.assertEqual(0, a.visits)
        self.assertEqual(10, b.visits)
        # Adding some submodels
        self.wait()
        ModelWithFK.objects.create(name='1', link=a, visits=30)
        ModelWithFK.objects.create(name='2', link=a, visits=2)
        self.synchronize()
        self.assertRemoteCount(1, PkModelWithSkip)
        a = PkModelWithSkip.objects.get(name='James')
        b = PkModelWithSkip.objects.db_manager(REMOTE).get(name='James')
        self.assertEqual(32, a.visits)
        self.assertEqual(42, b.visits)
        self.assertEqual(77, b.cash)  # No change in cash
        # Change
        self.wait()
        m2 = ModelWithFK.objects.get(name='2')
        m2.visits = 37
        m2.save()
        self.synchronize()
        a = PkModelWithSkip.objects.get(name='James')
        b = PkModelWithSkip.objects.db_manager(REMOTE).get(name='James')
        self.assertEqual(67, a.visits)
        self.assertEqual(77, b.visits)
        self.assertEqual(77, b.cash)  # Still no change in cash

    def _test_signals_scenario(self, handler, expected):
        """Scenario from README."""
        A.objects.create()
        self.synchronize()
        b = A.objects.db_manager(REMOTE).all()[0]
        b.foo = 42
        b.save()
        self.wait()
        post_save.connect(handler, sender=X, dispatch_uid='update_bar')
        X.objects.create(name='X')
        self.synchronize()
        a = A.objects.all()[0]
        b = A.objects.db_manager(REMOTE).all()[0]
        self.assertEqual(2, a.bar)  # handler was invoked
        self.assertEqual(2, b.bar)  # handler was invoked
        self.assertEqual(1, a.foo)
        self.assertEqual(expected, b.foo)
        post_save.disconnect(sender=X, dispatch_uid='update_bar')

    def test_signals_bad_scenario(self):
        """Demonstrate bad scenario from README."""
        self._test_signals_scenario(update_bar_bad, 1)  # BAD RESULT

    def test_signals_good_scenario(self):
        """Demonstrate solution for scenario from README (disable log)."""
        self._test_signals_scenario(update_bar_good_dis, 42)  # GOOD RESULT

    def test_signals_alternative_good_scenario(self):
        """Demonstrate solution for scenario from README (use update)."""
        self._test_signals_scenario(update_bar_good_upd, 42)  # GOOD RESULT


class M2MSynchroTests(SynchroTests):
    """Cover many2many relations tests."""

    def test_natural_manager(self):
        """Test if natural manager can be instantiated when using M2M."""
        test = M2mModelWithKey.objects.create()
        obj = M2mAnother.objects.create()
        obj.m2m.add(test)  # this would fail if NaturalManager could not be instantiated
        self.assertEqual(test.pk, obj.m2m.get_by_natural_key(1).pk)

    def test_simple_m2m(self):
        """Test if m2m field is synced properly."""
        test = M2mModelWithKey.objects.create()
        a = M2mAnother.objects.create()

        # add
        a.m2m.add(test)
        self.synchronize()
        self.assertRemoteCount(1, M2mAnother)
        self.assertRemoteCount(1, M2mModelWithKey)
        b = M2mAnother.objects.db_manager(REMOTE).all()[0]
        k = M2mModelWithKey.objects.db_manager(REMOTE).all()[0]
        self.assertEqual(1, b.m2m.count())
        self.assertEqual(1, k.r_m2m.count())
        b_k = b.m2m.all()[0]
        self.assertEqual(b_k.pk, k.pk)
        self.assertEqual(b_k.foo, k.foo)

        # clear
        self.wait()
        a.m2m.clear()
        self.synchronize()
        self.assertEqual(0, b.m2m.count())
        self.assertEqual(0, k.r_m2m.count())

        # reverse add
        self.wait()
        a2 = M2mAnother.objects.create(bar=2)
        test.r_m2m.add(a, a2)
        self.synchronize()
        self.assertRemoteCount(2, M2mAnother)
        self.assertRemoteCount(1, M2mModelWithKey)
        b2 = M2mAnother.objects.db_manager(REMOTE).filter(bar=2)[0]
        self.assertEqual(1, b.m2m.count())
        self.assertEqual(1, b2.m2m.count())
        self.assertEqual(2, k.r_m2m.count())

        # reverse remove
        self.wait()
        test.r_m2m.remove(a)
        self.synchronize()
        self.assertRemoteCount(2, M2mAnother)
        self.assertRemoteCount(1, M2mModelWithKey)
        self.assertEqual(0, b.m2m.count())
        self.assertEqual(1, b2.m2m.count())
        self.assertEqual(1, k.r_m2m.count())

        # reverse clear
        self.wait()
        test.r_m2m.clear()
        self.synchronize()
        self.assertRemoteCount(2, M2mAnother)
        self.assertRemoteCount(1, M2mModelWithKey)
        self.assertEqual(0, b.m2m.count())
        self.assertEqual(0, b2.m2m.count())
        self.assertEqual(0, k.r_m2m.count())

    def test_intermediary_m2m(self):
        """Test if m2m field with explicit intermediary is synced properly."""
        test = M2mNotExplicitlySynced.objects.create(foo=77)
        key = M2mModelWithKey.objects.create()
        a = M2mModelWithInter.objects.create()
        M2mIntermediate.objects.create(with_key=key, with_inter=a, cash=42, extra=test)
        self.assertEqual(1, a.m2m.count())
        self.assertEqual(1, key.r_m2m_i.count())
        self.synchronize()
        self.assertRemoteCount(1, M2mNotExplicitlySynced)
        self.assertRemoteCount(1, M2mModelWithKey)
        self.assertRemoteCount(1, M2mModelWithInter)
        b = M2mModelWithInter.objects.db_manager(REMOTE).all()[0]
        k = M2mModelWithKey.objects.db_manager(REMOTE).all()[0]
        self.assertEqual(1, b.m2m.count())
        self.assertEqual(1, k.r_m2m_i.count())
        b_k = b.m2m.all()[0]
        self.assertEqual(b_k.pk, k.pk)
        self.assertEqual(b_k.foo, k.foo)
        self.assertEqual(1, b.m2m.all()[0].foo)
        # intermediary
        self.assertRemoteCount(1, M2mIntermediate)
        inter = M2mIntermediate.objects.db_manager(REMOTE).all()[0]
        self.assertEqual(42, inter.cash)
        self.assertEqual(77, inter.extra.foo)  # check if extra FK model get synced

        # changing
        self.wait()
        key2 = M2mModelWithKey.objects.create(foo=42)
        a.m2m.clear()
        M2mIntermediate.objects.create(with_key=key2, with_inter=a, cash=77, extra=test)
        self.synchronize()
        self.assertRemoteCount(1, M2mNotExplicitlySynced)
        self.assertRemoteCount(2, M2mModelWithKey)
        self.assertRemoteCount(1, M2mModelWithInter)
        b = M2mModelWithInter.objects.db_manager(REMOTE).all()[0]
        self.assertEqual(1, b.m2m.count())
        self.assertEqual(42, b.m2m.all()[0].foo)
        # intermediary
        self.assertRemoteCount(1, M2mIntermediate)
        inter = M2mIntermediate.objects.db_manager(REMOTE).all()[0]
        self.assertEqual(77, inter.cash)

        # intermadiate change
        self.wait()
        inter = M2mIntermediate.objects.all()[0]
        inter.cash = 1
        inter.save()
        self.synchronize()
        # No changes here
        self.assertRemoteCount(1, M2mNotExplicitlySynced)
        self.assertRemoteCount(2, M2mModelWithKey)
        self.assertRemoteCount(1, M2mModelWithInter)
        b = M2mModelWithInter.objects.db_manager(REMOTE).all()[0]
        self.assertEqual(1, b.m2m.count())
        self.assertEqual(42, b.m2m.all()[0].foo)
        # Still one intermediary
        self.assertRemoteCount(1, M2mIntermediate)
        inter = M2mIntermediate.objects.db_manager(REMOTE).all()[0]
        self.assertEqual(1, inter.cash)


        # Tricky: clear from other side of relation.
        self.wait()
        key2.r_m2m_i.clear()
        self.synchronize()
        b = M2mModelWithInter.objects.db_manager(REMOTE).all()[0]
        self.assertEqual(0, b.m2m.count())
        self.assertRemoteCount(0, M2mIntermediate)

    def test_self_m2m(self):
        """Test if m2m symmetrical field is synced properly."""
        test = M2mSelf.objects.create(foo=42)
        a = M2mSelf.objects.create(foo=1)

        # add
        a.m2m.add(test)
        self.synchronize()
        self.assertRemoteCount(2, M2mSelf)
        b = M2mSelf.objects.db_manager(REMOTE).get(foo=1)
        k = M2mSelf.objects.db_manager(REMOTE).get(foo=42)
        self.assertEqual(1, b.m2m.count())
        self.assertEqual(1, k.m2m.count())
        b_k = b.m2m.all()[0]
        self.assertEqual(b_k.pk, k.pk)
        self.assertEqual(b_k.foo, k.foo)

        # clear
        self.wait()
        a.m2m.clear()
        self.synchronize()
        self.assertEqual(0, b.m2m.count())
        self.assertEqual(0, k.m2m.count())



from datetime import datetime

from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ObjectDoesNotExist
from django.core.management.base import BaseCommand, CommandError
from django.db import transaction

from synchro.models import Reference, ChangeLog, DeleteKey, options as app_options
from synchro.models import ADDITION, CHANGE, DELETION, M2M_CHANGE
from synchro.settings import REMOTE, LOCAL


def get_object_for_this_type_using(self, using, **kwargs):
    return self.model_class()._default_manager.using(using).get(**kwargs)
ContentType.get_object_for_this_type_using = get_object_for_this_type_using


def find_ref(ct, id):
    """
    Retrieves referenced remote object. Also deletes invalid reference.

    Returns (remote, reference) or (None, None).
    """
    try:
        ref = Reference.objects.get(content_type=ct, local_object_id=id)
        try:
            rem = ct.get_object_for_this_type_using(REMOTE, pk=ref.remote_object_id)
            return rem, ref
        except ObjectDoesNotExist:
            ref.delete()
            return None, None
    except Reference.DoesNotExist:
        return None, None


def find_natural(ct, loc, key=None):
    """Tries to find remote object for specified natural key or loc.natural_key."""
    try:
        key = key or loc.natural_key()
        model = ct.model_class()
        return model.objects.db_manager(REMOTE).get_by_natural_key(*key)
    except (AttributeError, ObjectDoesNotExist):
        return None


def is_remote_newer(loc, rem):
    try:
        loc_ct = ContentType.objects.get_for_model(loc)
        rem_ct = ContentType.objects.db_manager(REMOTE).get_for_model(rem)
        loc_time = (ChangeLog.objects.filter(content_type=loc_ct, object_id=loc.pk)
                    .order_by('-date')[0].date)
        rem_time = (ChangeLog.objects.filter(content_type=rem_ct, object_id=rem.pk)
                    .order_by('-date').using(REMOTE)[0].date)
        return rem_time >= loc_time
    except (ObjectDoesNotExist, IndexError):
        return False


def save_with_fks(ct, obj, new_pk):
    """
    Saves object in REMOTE, ensuring that every of it fk is present in REMOTE.
    Many-to-many relations are handled separately.
    """
    old_id = obj.pk
    obj._state.db = REMOTE

    fks = (f for f in obj._meta.fields if f.rel)
    for f in fks:
        fk_id = f.value_from_object(obj)
        if fk_id is not None:
            fk_ct = ContentType.objects.get_for_model(f.rel.to)
            rem, _ = ensure_exist(fk_ct, fk_id)
            f.save_form_data(obj, rem)

    obj.pk = new_pk
    obj.save(using=REMOTE)
    r, n = Reference.objects.get_or_create(content_type=ct, local_object_id=old_id,
                                           defaults={'remote_object_id': obj.pk})
    if not n and r.remote_object_id != obj.pk:
        r.remote_object_id = obj.pk
        r.save()

M2M_CACHE = {}


def save_m2m(ct, obj, remote):
    """Synchronize m2m fields from obj to remote."""
    model_name = obj.__class__

    if model_name not in M2M_CACHE:
        # collect m2m fields information: both direct and reverse
        res = {}
        for f in obj._meta.many_to_many:
            me = f.m2m_field_name()
            he_id = '%s_id' % f.m2m_reverse_field_name()
            res[f.attname] = (f.rel.to, f.rel.through, me, he_id)
        for rel in obj._meta.get_all_related_many_to_many_objects():
            f = rel.field
            if rel.get_accessor_name() is None:
                # In case of symmetrical relation
                continue
            me = f.m2m_reverse_field_name()
            he_id = '%s_id' % f.m2m_field_name()
            res[rel.get_accessor_name()] = (rel.model, f.rel.through, me, he_id)
        M2M_CACHE[model_name] = res

    _m2m = {}

    # handle m2m fields
    for f, (to, through, me, he_id) in M2M_CACHE[model_name].iteritems():
        fk_ct = ContentType.objects.get_for_model(to)
        out = []
        if through._meta.auto_created:
            for fk_id in getattr(obj, f).using(LOCAL).values_list('pk', flat=True):
                rem, _ = ensure_exist(fk_ct, fk_id)
                out.append(rem)
        else:
            # some intermediate model is used for this m2m
            inters = through.objects.filter(**{me: obj}).using(LOCAL)
            for inter in inters:
                ensure_exist(fk_ct, getattr(inter, he_id))
                out.append(inter)
        _m2m[f] = not through._meta.auto_created, out

    for f, (intermediary, out) in _m2m.iteritems():
        if not intermediary:
            setattr(remote, f, out)
        else:
            getattr(remote, f).clear()
            for inter in out:
                # we don't need to set any of objects on inter. References will do it all.
                ct = ContentType.objects.get_for_model(inter)
                save_with_fks(ct, inter, None)


def create_with_fks(ct, obj, pk):
    """Performs create, but firstly disables synchro of some user defined fields (if any)"""
    skip = getattr(obj, 'SYNCHRO_SKIP', ())
    raw = obj.__class__()
    for f in skip:
        setattr(obj, f, getattr(raw, f))
    return save_with_fks(ct, obj, pk)


def change_with_fks(ct, obj, rem):
    """Performs change, but firstly disables synchro of some user defined fields (if any)"""
    skip = getattr(obj, 'SYNCHRO_SKIP', ())
    for f in skip:
        setattr(obj, f, getattr(rem, f))
    return save_with_fks(ct, obj, rem.pk)


def ensure_exist(ct, id):
    """
    Ensures that remote object exists for specified ct/id. If not, create it.
    Returns remote object and reference.
    """
    obj = ct.get_object_for_this_type(pk=id)
    rem, ref = find_ref(ct, obj.pk)
    if rem is not None:
        return rem, ref
    rem = find_natural(ct, obj)
    if rem is not None:
        ref = Reference.create(content_type=ct, local_object_id=id, remote_object_id=rem.pk)
        return rem, ref
    return perform_add(ct, id)


def perform_add(ct, id, log=None):
    obj = ct.get_object_for_this_type(pk=id)
    rem = find_natural(ct, obj)
    if rem is not None:
        if not is_remote_newer(obj, rem):
            change_with_fks(ct, obj, rem)
            rem = obj
    else:
        new_pk = None if obj._meta.has_auto_field else obj.pk
        create_with_fks(ct, obj, new_pk)
        rem = obj
    ref, _ = Reference.objects.get_or_create(content_type=ct, local_object_id=id,
                                             remote_object_id=rem.pk)
    return rem, ref


def perform_chg(ct, id, log=None):
    obj = ct.get_object_for_this_type(pk=id)
    rem, ref = find_ref(ct, obj.pk)
    if rem is not None:
        return change_with_fks(ct, obj, rem)
    rem = find_natural(ct, obj)
    if rem is not None:
        return change_with_fks(ct, obj, rem)
    perform_add(ct, id)


def perform_del(ct, id, log):
    rem, ref = find_ref(ct, id)
    if rem is not None:
        return rem.delete()
    try:
        raw_key = log.deletekey.key
        key = eval(raw_key)
        rem = find_natural(ct, None, key)
        if rem is not None:
            rem.delete()
    except DeleteKey.DoesNotExist:
        pass


def perform_m2m(ct, id, log=None):
    obj = ct.get_object_for_this_type(pk=id)
    rem, ref = find_ref(ct, obj.pk)
    if rem is not None:
        return save_m2m(ct, obj, rem)
    rem = find_natural(ct, obj)
    if rem is not None:
        return save_m2m(ct, obj, rem)
    rem, _ = perform_add(ct, id)
    return save_m2m(ct, obj, rem)


ACTIONS = {
    ADDITION:   perform_add,
    CHANGE:     perform_chg,
    DELETION:   perform_del,
    M2M_CHANGE: perform_m2m,
}


class Command(BaseCommand):
    args = ''
    help = '''Perform synchronization.'''

    @transaction.commit_on_success
    @transaction.commit_on_success(using=REMOTE)
    def handle(self, *args, **options):
        if REMOTE is None:
            # Because of BaseCommand bug (#18387, fixed in Django 1.5), we cannot use CommandError
            # in tests. Hence this hook.
            exception_class = options.get('exception_class', CommandError)
            raise exception_class('No REMOTE database specified in settings.')

        since = app_options.last_check
        last_time = datetime.now()
        logs = ChangeLog.objects.filter(date__gt=since).select_related().order_by('date', 'pk')

        # Don't synchronize if object should be added/changed and later deleted;
        to_del = {}
        for log in logs:
            if log.action == DELETION:
                to_del[(log.content_type, log.object_id)] = log.date

        for log in logs:
            last_time = log.date
            del_time = to_del.get((log.content_type, log.object_id))
            if last_time == del_time and log.action == DELETION:
                ACTIONS[log.action](log.content_type, log.object_id, log)
                # delete record so that next actions with the same time can be performed
                del to_del[(log.content_type, log.object_id)]
            if del_time is None or last_time > del_time:
                ACTIONS[log.action](log.content_type, log.object_id, log)

        if len(logs):
            app_options.last_check = last_time
            msg = 'Synchronization performed successfully.'
        else:
            msg = 'No changes since last synchronization.'

        if options['verbosity'] > 0:
            self.stdout.write('%s\n' % msg)



from functools import wraps

from django.db.models.signals import post_save, post_delete, m2m_changed


def synchro_connect():
    from models import save_changelog_add_chg, save_changelog_del, save_changelog_m2m
    post_save.connect(save_changelog_add_chg, dispatch_uid='synchro_add_chg')
    post_delete.connect(save_changelog_del, dispatch_uid='synchro_del')
    m2m_changed.connect(save_changelog_m2m, dispatch_uid='synchro_m2m')


def synchro_disconnect():
    post_save.disconnect(dispatch_uid='synchro_add_chg')
    post_delete.disconnect(dispatch_uid='synchro_del')
    m2m_changed.disconnect(dispatch_uid='synchro_m2m')


class DisableSynchroLog(object):
    def __enter__(self):
        synchro_disconnect()

    def __exit__(self, *args, **kwargs):
        synchro_connect()
        return False


def disable_synchro_log(f):
    @wraps(f)
    def inner(*args, **kwargs):
        with DisableSynchroLog():
            return f(*args, **kwargs)
    return inner



from django.contrib.admin.models import ADDITION, CHANGE, DELETION
from django.contrib.contenttypes.models import ContentType
from django.contrib.contenttypes import generic
from django.db import models
import dbsettings

#noinspection PyUnresolvedReferences
import settings  # in order to validate

M2M_CHANGE = 4

ACTIONS = (
    (ADDITION, 'Add'),
    (CHANGE, 'Change'),
    (DELETION, 'Delete'),
    (M2M_CHANGE, 'M2m Change'),
)


class SynchroSettings(dbsettings.Group):
    last_check = dbsettings.DateTimeValue('Last synchronization')
options = SynchroSettings()


class Reference(models.Model):
    content_type = models.ForeignKey(ContentType)
    local_object_id = models.CharField(max_length=20)
    remote_object_id = models.CharField(max_length=20)

    class Meta:
        unique_together = ('content_type', 'local_object_id')


class ChangeLog(models.Model):
    content_type = models.ForeignKey(ContentType)
    object_id = models.CharField(max_length=20)
    object = generic.GenericForeignKey()
    date = models.DateTimeField(auto_now=True)
    action = models.PositiveSmallIntegerField(choices=ACTIONS)

    def __unicode__(self):
        return u'ChangeLog for %s (%s)' % (unicode(self.object), self.get_action_display())


class DeleteKey(models.Model):
    changelog = models.OneToOneField(ChangeLog)
    key = models.CharField(max_length=200)


def delete_redundant_change(cl):
    """
    Takes ChangeLog instance as argument and if previous ChangeLog for the same object
    has the same type, deletes it.
    It ensures that if several object's changes were made one-by-one, only one ChangeLog is stored
    afterwards.
    """
    cls = (ChangeLog.objects.filter(content_type=cl.content_type, object_id=cl.object_id)
           .exclude(pk=cl.pk).order_by('-date', '-pk'))
    if len(cls) > 0 and cls[0].action == cl.action:
        cls[0].delete()


def save_changelog_add_chg(sender, instance, created, using, **kwargs):
    if sender in settings.MODELS and using == settings.LOCAL:
        if created:
            ChangeLog.objects.create(object=instance, action=ADDITION)
        else:
            cl = ChangeLog.objects.create(object=instance, action=CHANGE)
            delete_redundant_change(cl)
    elif sender in settings.INTER_MODELS and using == settings.LOCAL:
        rel = settings.INTER_MODELS[sender]
        # It doesn't matter if we select forward or reverse object here; arbitrary choose forward
        real_instance = getattr(instance, rel.field.m2m_field_name())
        cl = ChangeLog.objects.create(object=real_instance, action=M2M_CHANGE)
        delete_redundant_change(cl)


def save_changelog_del(sender, instance, using, **kwargs):
    if sender in settings.MODELS and using == settings.LOCAL:
        cl = ChangeLog.objects.create(object=instance, action=DELETION)
        try:
            k = repr(instance.natural_key())
            DeleteKey.objects.create(changelog=cl, key=k)
        except AttributeError:
            pass


def save_changelog_m2m(sender, instance, model, using, action, **kwargs):
    if ((model in settings.MODELS or instance.__class__ in settings.MODELS)
            and action.startswith('post') and using == settings.LOCAL):
        cl = ChangeLog.objects.create(object=instance, action=M2M_CHANGE)
        delete_redundant_change(cl)


# start logging
from signals import synchro_connect
synchro_connect()



from cStringIO import StringIO

from django.contrib.admin.views.decorators import staff_member_required
from django.contrib import messages
from django.template.response import TemplateResponse
from django.core.management import call_command

from models import options


@staff_member_required
def synchro(request):
    if 'synchro' in request.POST:
        so = StringIO()
        try:
            call_command('synchronize', stdout=so)
            messages.add_message(request, messages.INFO, so.getvalue())
        except Exception as e:
            msg = 'An error occured: %s (%s)' % (str(e), e.__class__.__name__)
            messages.add_message(request, messages.ERROR, msg)
        finally:
            so.close()
    return TemplateResponse(request, 'synchro.html', {'last': options.last_check})



#!/usr/bin/env python
from setuptools import setup, find_packages

setup(
    name='django-synchro',
    description='Django app for database data synchronization.',
    long_description=open('README.rst').read(),
    version='0.5.1',
    author='Jacek Tomaszewski',
    author_email='jacek.tomek@gmail.com',
    url='https://github.com/zlorf/django-synchro',
    license='MIT',
    install_requires=(
        'django-dbsettings>=0.1',
        'django>=1.4',
    ),
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Framework :: Django',
    ],
    packages=find_packages(),
    include_package_data = True,
)



from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.db.models.loading import get_app, get_models, get_model, load_app


def get_all_models(app):
    try:
        app_mod = load_app(app) # First try full path
    except ImportError:
        app_mod = get_app(app) # Then try just app_label
    return get_models(app_mod)


def gel_listed_models(app, l):
    def parse(model):
        m = get_model(app, model)
        if m is None:
            raise ImproperlyConfigured(
                'SYNCHRO_MODELS: Model %s not found in %s app.' % (model, app))
        return m
    return map(parse, l)


def parse_models(l):
    res = []
    for entry in l:
        if len(entry) == 1:
            entry = entry[0]
        if type(entry) == str:
            res.extend(get_all_models(entry))
        else:
            app = entry[0]
            res.extend(gel_listed_models(app, entry[1:]))
    return res


def get_intermediary(models):
    res = {}
    for model in models:
        res.update((m2m.rel.through, m2m.related) for m2m in model._meta.many_to_many
                   if not m2m.rel.through._meta.auto_created)
    return res


MODELS = parse_models(getattr(settings, 'SYNCHRO_MODELS', ()))
# Since user-defined m2m intermediary objects don't send m2m_changed signal, we need to listen to
# those models.
INTER_MODELS = get_intermediary(MODELS)
REMOTE = getattr(settings, 'SYNCHRO_REMOTE', None)
LOCAL = 'default'

if REMOTE is None:
    if not hasattr(settings, 'SYNCHRO_REMOTE'):
        import  warnings
        warnings.warn('SYNCHRO_REMOTE not specified. Synchronization is disabled.', RuntimeWarning)
elif REMOTE not in settings.DATABASES:
    raise ImproperlyConfigured('SYNCHRO_REMOTE invalid - no such database: %s.' % REMOTE)
